<?php 

namespace Gogs\API\Request {

    /** 
     * Stores data and methods related to a single repository.
     *
     * By now the following are supported:
     *
     *  * GET `/repos/username/reponame`
     *  * POST `/user/repos`
     *  * POST `/admin/user/username/repos`
     *  * POST `/org/orgname/repos`
     *  * DELETE `/repos/username/reponame`
     *
     * @author Joachim M. Giaever (joachim[]giaever.org)
     */
    final class Repo extends Base {
        
        public $repo_id;
        public $repo_owner;
        public $repo_name;
        public $repo_full_name;
        public $repo_description;
        public $repo_private;
        public $repo_fork;
        public $repo_parent;
        public $repo_empty;
        public $repo_mirror;
        public $repo_size;
        public $repo_html_url;
        public $repo_ssh_url;
        public $repo_clone_url;
        public $repo_website;
        public $repo_stars_count;
        public $repo_forks_count;
        public $repo_watchers_count;
        public $repo_open_issues_count;
        public $repo_default_branch;
        public $repo_created_at;
        public $repo_updated_at;
        public $repo_permissions;
        public $repo_admin;
        public $repo_push;
        public $repo_pull;

        /**
         * Initialize a repo object.
         *
         * Note that the owner can also be an Org (organization),
         * or any other class that inherits a user.
         *
         * @param string $api_url The api-url
         * @param string $api_token The api-token
         * @param User $owner The owner of the repo
         * @param string $name The repo name
         */
        public function __construct(string $api_url, string $api_token, User $owner = null, string $name = null) {
            parent::__construct($api_url, $api_token);
            $this->owner = $owner;
            $this->name = $name;
        }

        /** 
         * @see Base
         * @throws Exception\RequestErrorException on missing repo or user data
         */
        protected function set_scope(string $method) {
            switch ($method) {
            case "create":
                if (!$this->owner->authenticated() && empty($this->owner->username))
                    throw new Exception\RequestErrorException("Missing userdata of unauthorized user 'username'");

                if ($this->owner instanceof Org)
                    $this->scope = "/org/" . $this->owner->username . "/repos";
                elseif ($this->owner->authenticated())
                    $this->scope = "/user/repos";
                else
                    $this->scope = "/admin/users/" . $this->owner->username . "/repos";
                break;
            case "delete":
                if (empty($this->owner->username))
                    throw new Exception\RequestErrorException("Missing userdata 'username'");

                $this->scope = "/repos/" . $this->owner->username . "/" . $this->name;
                break;
            case "get":
                if (empty($this->owner->username) && empty($this->full_name))
                    throw new Exception\RequestErrorException("Missing userdata 'username' and/or 'full_name'");

                $this->scope = "/repos/" . ($this->owner ? $this->owner->username . "/" . $this->name : $this->full_name);
                break;
            default:
                return false;
            }

            return true;
        }

        /** 
         * @see Base
         */
        protected function json_set_property($obj) {
            foreach($obj as $key => $val) {
                if ($this->property_exists($key)) {
                    switch ($key) {
                    case "owner":
                        if (!$this->owner) {
                            $user = new User($this->url, $this->token);
                            $user->json_set_property($val);
                            $this->{$key} = $user;
                        }
                        break;
                    default:
                        $this->{$key} = $val;
                    }
                }
            }
            $this->loaded = true;

            return true;
        }

        /** 
         * @see Base
         */
        public function search(string $q) {
            $searchable = sprintf("%s %s", $this->name, $this->description);

            return stripos($searchable, $q) !== false;
        }

        /** 
         * Create a new repo
         *
         * Valid paramters:
         *
         *  1. name
         *  2. description
         *  3. private (default: false)
         *  4. auto_init (default: false)
         *  5. gitignore
         *  6. license
         *  7. readme (default: "Default")
         *   
         *   This reflects the API v1 documentation, but is in an order
         *   where the required fields are first.
         *
         * @param ...$args The parameter values
         * @return bool
         */
        public function create(...$args) {

            $params = array(
                "name" => isset($args[0]) && is_string($args[0]) ? $args[0] : null,
                "description" => isset($args[1]) && is_string($args[1]) ? $args[1] : null,
                "private" => isset($args[2]) && is_bool($args[2]) ? $args[2] : false,
                "auto_init" => isset($args[3]) && is_bool($args[3]) ? $args[3] : false,
                "gitignores" => isset($args[4]) && is_string($args[4]) ? $args[4] : null,
                "licence" => isset($args[5]) && is_string($args[5]) ? $args[5] : null,
                "readme" => isset($args[6]) && is_string($args[6]) ? $args[6] : "Default"
            );

            $params = array_filter($params, function($val) {
                return $val != null;
            });

            parent::create($params);
        }
    }

    }
?>
